<?php

/* smtp.php: smtp routines
*
    Copyright (C) 2002  Hastymail Development group

    This program is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; Version 2
    See the COPYING file for complete information.

* $Id: smtp.php,v 1.77 2003/10/10 19:46:25 sailfrog Exp $
*/

require_once('general.php');

/* the main function for sending mail, is called from compose.php
   $to              = to addresses string
   $cc              = cc addresses string
   $bcc             = bcc addresses string
   $subject         = subject string
   $message         = message string
   $from            = from address string
   $reply_inreplyto = inreplyto header value for replies
   $reply_messid    = the message id of the message replied to
   $reply_refs      = the references header value for replies
   $reply           = true if we are replying to a message
   $config          = the configuration array
   returns false on failure and true on success
*/
function hm_send_mail($to, $cc, $bcc, $subject, $message, $from, $reply_inreplyto,
                      $reply_messid, $reply_refs, $reply, $config, $config_file,
                      $bounce=false, $version, $hasty_key, $imap_stream) {

    $str = $_SESSION['str'];
    $hostname = preg_replace("/^http(s){0,1}:\/\//", '', $config['hostname']);
    $hostname = preg_replace("/:\d{1,}$/", '', $hostname);
    $domain = $config['domain'];
    $server = $config['smtp_server'];
    $no_reply_refs = false;
    if ($reply) {
        if ($reply_messid == '') {
            $no_reply_refs = true;
        }
        else {
            $irt = $reply_messid;
            if ($reply_refs != '') {
                $refs = $reply_refs.' '.$reply_messid;
            }
            else {
                if ($reply_inreplyto != '') {
                    $refs = $reply_inreplyto.' ';
                }
                $refs = $reply_messid;
            }
            $refs = preg_replace("/(\t|\r\n)/", ' ', $refs);
            $refs = preg_replace("/\s{2,}/", ' ', $refs);
            $refs_array = split(' ', $refs); 
            $refs = implode("\r\n            ", $refs_array);
        }
    }
    if (!$bounce) {
        $from_address = $_SESSION['settings'][$from];
        $index = substr($from, (strlen($from)-1));
        if (isset($_SESSION['settings']['from_name_'.$index])) {
            $from_name = '"'.$_SESSION['settings']['from_name_'.$index].'"';
        }
        else {
            $from_name = '';
        }
    }
    else {
        $from_address = $from;  
        $from_name = '';
    }
    
    $lost = 
'<table align="center" class="tables3">
   <tr>\n".
      <td align="center">
      <font class="font2"><small>
      '.$str[179].'
      </small></font>
      </td>
   </tr>
</table><br>'."\n";

    /* connect to the SMTP server */
    if (!$smtp_stream = hm_connect_smtp($server)) {
        return false;
    }
    
    /* say hello */
    if (!hm_smtp_helo($smtp_stream, $hostname, $lost)) {
        if ($smtp_stream) {
            fclose($smtp_stream);
        }
        return false;
    }
    
    /* smtp auth */
    if (isset($config['smtp_auth']) && $config['smtp_auth'] == 'cram-md5' ||
        $config['smtp_auth'] == 'plain' || $config['smtp_auth'] == 'login') {
        if(!hm_auth_smtp($smtp_stream, $config['smtp_auth'], $lost, $config_file,
            $hasty_key)) {
            if ($smtp_stream) {
                fclose($smtp_stream);
            }
            return false;
        }
    }
       
    if (!hm_smtp_mail($smtp_stream, $lost, $from_address)) {
        if ($smtp_stream) {
            fclose($smtp_stream);
        }
        return false;
    }
    
    /* parse To: addresses */
    $to_clean = preg_replace("/(\"[^\"]*\"|<|>|\([^\)]*\))/", "", $to);
    $to_array = preg_split("/(\s*,\s*|\s*;\s*)/", $to_clean);
    $i = 0;
    foreach ($to_array as $to_addy) {
        $to_addy = trim($to_addy, ",; ");
        if ($to_addy != '') {
            if (!hm_smtp_rcpt($smtp_stream, $to_addy, $lost)) {
                $i++;
            }
        }
    }
    if ($i == count($to_array)) {
        if ($smtp_stream) {
            fclose($smtp_stream);
        }
        return false;
    }

    /* parse Cc: addresses */
    if ($cc != '') {
        $cc = preg_replace("/(\"[^\"]*\"|<|>|\([^\)]*\))/", "", $cc);
        $cc_array = preg_split("/(\s*,\s*|\s*;\s*)/", $cc);
        foreach ($cc_array as $cc_addy) {
            $cc_addy = trim($cc_addy, ",; ");
            if ($cc_addy != '') {
                if (!hm_smtp_rcpt($smtp_stream, $cc_addy, $lost)) {
                }
            }
        }
    }

    /* parse Bcc: addresses */
    if ($bcc != '') {
        $bcc = preg_replace("/(\"[^\"]*\"|<|>|\([^\)]*\))/", "", $bcc);
        $bcc_array = preg_split("/(\s*,\s*|\s*;\s*)/", $bcc);
        foreach ($bcc_array as $bcc_addy) {
            $bcc_addy = trim($bcc_addy, ",; ");
            if ($bcc_addy != '') {
                if (!hm_smtp_rcpt($smtp_stream, $bcc_addy, $lost)) {
                }
            }
        }
    }

    /* check fields for required encodings */
    if (hm_special_char_check($from_name)) {
        $from_name = '=?'.$str['charset'].'?B?'.base64_encode($from_name).'?=';
    }
    if (hm_special_char_check($subject)) {
        $subject = '=?'.$str['charset'].'?B?'.base64_encode($subject).'?=';
    }
    if (hm_special_char_check($message)) {
        $content_charset = $str['charset'];
    }
    else {
        $content_charset = 'us-ascii';
    }

    /* build message headers and body to send to the SMTP server */
    $smtp_message  = 'Message-ID: <'.md5(uniqid(rand(),1))."@$domain>\r\n";
    $smtp_message .= 'Date: '.date("r")."\r\n";
    $smtp_message .= "From: $from_name <$from_address>\r\n";
    $smtp_message .= "Subject: $subject\r\n";
    $smtp_message .= "To: $to\r\n";
    $smtp_message .= "Reply-to: <$from_address>\r\n";
    if ($cc != '') {
        $smtp_message .= "Cc: $cc\r\n";
    }
    if ($reply && !$no_reply_refs) {
        $smtp_message .= "In-Reply-To: ".str_replace(array('&gt;', '&lt;'), 
                                         array('>', '<'), $irt)."\r\n";
        $smtp_message .= "References: ".str_replace(array('&gt;', '&lt;'), 
                                        array('>', '<'), $refs)."\r\n";
    }
    $size = 0;
    $bndry = '';
    $smtp_message .= "X-Mailer: Hastymail $version\r\n";
    $smtp_message .= "MIME-Version: 1.0\r\n";
    
    /* calculations and headers for multi-part messages */
    if (isset($_SESSION['attachments']) && count($_SESSION['attachments']) > 0) {
        $bndry = md5(uniqid(rand(),1));
        foreach($_SESSION['attachments'] as $array) {
            $size += ($array['rawsize'] + 8 + strlen($array['header']) + strlen($bndry));
        }
        $size += (strlen($bndry) + 4);
        $smtp_message .= "Content-Type:multipart/mixed; boundary=\"$bndry\"\r\n\r\n";
        $bndry = "--".$bndry;
        $smtp_message .= "$bndry\r\n";
        $smtp_message .= "Content-Type: text/plain; charset=".$content_charset."\r\n";
        $smtp_message .= "Content-Transfer-Encoding: 8bit\r\n";
        $multi = true;
        
    }
    else {
        $smtp_message .= "Content-type:text/plain; charset=".$content_charset."\r\n";
        $multi = false;
    }

    /* properly format the body text */
    $lines = split("\n", $message);
    $new_lines = array();
    $new_imap_lines = array();
    foreach($lines as $string) {
        if (strlen($string) > 78) {
            $line = rtrim(implode("\r\n", hm_split_line($string, 76, false, true)),
                           "\n")."\r\n";
            $new_imap_lines[] = $line;
            $new_lines[] = preg_replace("/^\.\r\n/", "..\r\n", $line);
        }
        else {
            $line = rtrim($string, "\r\n")."\r\n";
            $new_imap_lines[] = $line;
            $new_lines[] = preg_replace("/^\.\r\n/", "..\r\n", $line);
        }
    }

    /* final formatting before the SMTP data command */
    $body = implode("", $new_lines);
    $imap_body = implode("", $new_imap_lines);
    $imap_message = $smtp_message."\r\n".$imap_body;
    $smtp_message .= "\r\n".$body;
    $size += strlen($imap_message);
    if (isset($_SESSION['forward_flag']) && $_SESSION['forward_flag'] == 'true') {
        unset($_SESSION['forward_flag']);
        $size -= 4;
    }
    if (strlen($imap_body) == 2) {
        $imap_message .= "\r\n\r\n";
        $size -= 4;
    }

    /* send the headers/body */
    if (!hm_smtp_data($smtp_stream, $smtp_message, $lost, $reply, $multi, $bndry)) {
        if (is_resource($smtp_stream)) {
            fclose($smtp_stream);
        }
        return false;
    }
    /* clean up */
    else {
        fputs($smtp_stream, "QUIT\r\n");
        fclose($smtp_stream);
        $config = $_SESSION['config'];
        hm_save_message($imap_message, $imap_stream, 'sent', $multi, $bndry, $size);
        hm_imap_logout($imap_stream); 
        return true;
    }
}

/* connect to the SMTP server 
   $server = smtp server address or hostname
   returns $smtp_stream, a connection to the smtp server on success
   or false on failure
*/
function hm_connect_smtp($server) {
    $str = $_SESSION['str'];
    if (isset($_SESSION['config']['smtp_port'])) {
        $port = $_SESSION['config']['smtp_port'];
    }
    else {
        $port = 25;
    }
    if (isset($_SESSION['config']['smtp_ssl']) && 
        $_SESSION['config']['smtp_ssl'] == 'ssl') {
        $server = 'ssl://'.$server;
    }
    elseif (isset($_SESSION['config']['smtp_ssl']) && 
        $_SESSION['config']['smtp_ssl'] == 'tls') {
        $server = 'tls://'.$server;
    }
    if ($smtp_stream = @fsockopen($server, $port, $errorno, $errorstr, 30)) {
        if ($helo = fgets($smtp_stream, 1024)) {
            if (preg_match("/^220/", $helo)) {
                return $smtp_stream;
            }
            else {
                echo '<table align="center" class="tables3">
   <tr>
      <td align="center">
      <font class="font2"><small>
      ',$str[180],':<br>
      ',$helo,'
      </small></font>
      </td>
   </tr>
</table><br>',"\n";
                return false;
            }
        }
        else {
                echo '<table align="center" class="tables3">
   <tr>
      <td align="center">
      <font class="font2"><small>
      ',$str[181],' 
      </small></font>
      </td>
   </tr>
</table><br>',"\n";
            return false;
        }
    }
    else {
                echo '<table align="center" class="tables3">
   <tr>
      <td align="center">
      <font class="font2"><small>
      ',$str[181],'
      </small></font>
      </td>
   </tr>
</table><br>',"\n";
        return false;
    }
}
    
/* sends the SMTP HELO command
   $smtp_stream = connection to the smtp server
   $hostname    = the hostname we are sending mail from
   returns true on success, false on failure
*/
function hm_smtp_helo($smtp_stream, $hostname) {
    $str = $_SESSION['str'];
    fputs($smtp_stream, "HELO $hostname\r\n");
    if ($response = fgets($smtp_stream, 1024)) {
        if (substr($response, 0, 4) == '250 ') {
            return true;
        }
        elseif (substr($response, 0, 4) == '250-') {
            while (substr($response, 0, 4) == '250-') {
                $response = fgets($smtp_stream, 1024);
            }
            return true;
        }
        else {
            echo '<table align="center" class="tables3">
  <tr>
      <td align="center">
      <font class="font2"><small>
      ',$str[182],':<br>
      ',$response,'
      </small></font>
      </td>
   </tr>
</table><br>',"\n";
            return false;
        }
        
    }
    else {
        echo $lost;
        return false;
    }
}

function hm_auth_smtp($smtp_stream, $auth, $lost, $config_file, $hasty_key) {
    $str = $_SESSION['str'];
    if ($auth != 'plain') {
        $command = 'AUTH '.strtoupper($auth)."\r\n";
    }
    else {
        $key = hmw_crypt($_SESSION['pass'], $config, 'decrypt', $hasty_key);
        $user = $_SESSION['user'];
        $command = 'AUTH PLAIN '.base64_encode($user."\0".$user."\0".$key)."\r\n";
    }
    fputs($smtp_stream, $command);
    if ($response = fgets($smtp_stream, 1024)) {
        if (substr($response, 0, 3) != '334' && $auth != 'plain') {
            echo '<table align="center" class="tables3">
  <tr>
      <td align="center">
      <font class="font2"><small>
      ',$str[234],':<br>
      ',$response,'
      </small></font>
      </td>
   </tr>
</table><br>',"\n";
            return false;
        }
        else {
            switch ($auth) {
                case 'cram-md5':
                    $challenge = base64_decode(substr($response, 4));
                    $user = $_SESSION['user'];
                    $key = hmw_crypt($_SESSION['pass'], $config, 'decrypt', $hasty_key); 
                    $key .= str_repeat(chr(0x00), (64-strlen($key)));
                    $ipad = str_repeat(chr(0x36), 64);
                    $opad = str_repeat(chr(0x5c), 64);
                    $digest = bin2hex(pack("H*", md5(($key ^ $opad).
                              pack("H*", md5(($key ^ $ipad).$challenge)))));
                    $challenge_response = base64_encode($user.' '.$digest);
                    fputs($smtp_stream, $challenge_response."\r\n");
                    $result = fgets($smtp_stream, 1024);
                    break;
                case 'login':
                    $key = hmw_crypt($_SESSION['pass'], $config, 'decrypt', $hasty_key);
                    $user = $_SESSION['user'];
                    fputs($smtp_stream, base64_encode($user)."\r\n");
                    $prompt = fgets($smtp_stream, 1024);
                    fputs($smtp_stream, base64_encode($key)."\r\n");
                    $result = fgets($smtp_stream, 1024);
                    break;
                case 'plain':
                    $result = $response;
                    break;
                default:
                   $result = $response;
                    break;
            }
            if (substr($result, 0, 3) != '235') {
                echo '<table align="center" class="tables3">
  <tr>
      <td align="center">
      <font class="font2"><small>
      ',$str[234],':<br>
      ',$response,'
      </small></font>
      </td>
   </tr>
</table><br>',"\n";
                return false;
            }
            else {
                return true;
            }
        }
    }
    else {
        echo $lost;
        return false;
    }
}

/* performs the SMTP MAIL command
   $smtp_stream = connection to the smtp server
   $lost        = some html to echo if our socket connection went away
   $from        = the senders email address
   returns true on success, false on failure
*/
function hm_smtp_mail ($smtp_stream, $lost, $from) {
    $str = $_SESSION['str'];
    fputs($smtp_stream, "MAIL FROM: <$from>\r\n");
    if ($response = fgets($smtp_stream, 1024)) {
        if (preg_match("/^250/", $response)) {
            return true;
        }
        else {
            echo '<table align="center" class="tables3">
   <tr>
      <td align="center">
      <font class="font2"><small>
      ',$str[183],'<br>
      ',$response,'
      </small></font>
      </td>
   </tr>
</table><br>',"\n";
            return false;
        }
    }
    else {
        echo $lost;
        return false;
    }
}

/* performs the SMTP RCPT command 
   $smtp_stream = connection to the smtp server
   $lost        = some html to echo if our socket connection went away
   $to          = a single to address
   returns true on success, false on failure
*/
function hm_smtp_rcpt ($smtp_stream, $to, $lost) {
    $str = $_SESSION['str'];
    //$dmpart = strrev(strrev(strstr($to, '@')));
    fputs($smtp_stream, "RCPT TO: <$to>\r\n");
    if ($response = fgets($smtp_stream, 1024)) {
        if (preg_match("/^25/", $response)) {
            return true;
        }
        else {
            echo "<table align=\"center\" class=\"tables3\">\n",
                 "   <tr>\n",
                 "      <td align=\"center\">\n",
                 "      <font class=\"font2\"><small>\n",
                 "      There was an error with the recipient $to,<br>\n",
                 "      The SMTP server said:<br>\n",
                        $response,
                 "      </small></font>\n",
                 "      </td>\n",
                 "   </tr>\n",
                 "</table><br>\n";
            return false;
        }
        
    }
    else {
        echo $lost;
        return false;
    }
}

/* sends the SMTP data (headers and body) 
   $smtp_stream = connection to the smtp server
   $lost        = some html to echo if our socket connection went away
   $message     = the message to send
   $reply       = true if the message is a reply
   returns true on success, false on failure
*/
function hm_smtp_data ($smtp_stream, $message, $lost, $reply, $multi, $bndry) {
    $str = $_SESSION['str'];
    fputs($smtp_stream, "DATA\r\n");
    if ($response = fgets($smtp_stream, 1024)) {
        if (preg_match("/^354/", $response)) {
            if ($multi) {
                if(!hm_write_data($smtp_stream, $message, $bndry)) {
                    echo $lost; 
                    return false;
                }
                else {
                    fputs($smtp_stream, "\r\n.\r\n");
                }
            }
            else {
                    fputs($smtp_stream, $message."\r\n.\r\n");
            }
            if ($final_response = fgets($smtp_stream, 1024)) {
                if (preg_match("/^250/", $final_response)) {
                    if (!$reply) {
                       echo '<table align="center" class="tables3">
   <tr>
      <td align="center">
      <font class="font2"><small>
      ',$str[184],'
      </small></font>
      </td>
   </tr>
</table><br>',"\n";
                    }
                    return true;
                }
                else {
                   echo '<table align="center" class="tables3">
   <tr>
      <td align="center">
      <font class="font2"><small>
      ',$str[185],':<br>
      ',$final_response,'
      </small></font>
      </td>
   </tr>
</table><br>',"\n";
                    return false;
                }
            }
            else {
                echo $lost; 
                return false;
            }
        }
        else {
            echo '<table align="center" class="tables3">
   <tr>
      <td align="center">
      <font class="font2"><small>
      ',$str[185],':<br>
      ',$response,'
      </small></font>
      </td>
   </tr>
</table><br>',"\n";
            return false;
        }
        
    }
    else {
        echo $lost;
        return false;
    }
}

/* checks for non-ascii chars.
    $string = string to check
   returns true if a char over ASCII code 127 is found
*/
function hm_special_char_check($string) {
    $result = false;
    $len = strlen($string);
    for ($i=0;$i<$len;$i++) {
        if (ord($string{$i}) > 127) {
            $result = true;
            break;
        }
    }
    return $result;
}
?>
